/*
 * coverage.c
 * $Revision: 1.1 $
 *
 * Copyright (c) 2004 David Holroyd
 * See the file COPYING for license terms
 *
 * http://www.badgers-in-foil.co.uk/projects/cvprof/
 */

#define NO_JVMDI_MACROS
#include <jvmdi.h>
#include <jni.h>
#include <stdlib.h>
#include "err.h"
#include "coverage.h"
#include "bitset.h"
#include "hash.h"
#include "hash_utils.h"

extern JVMDI_Interface_1 *jvmdi;
extern hash_table hash;

/*
 * Returns the length of this method, or -1 if the method length can't be
 * determined.  Note this will usually be greater than the number of
 * bytecodes, since the length includes both bytecodes and 'immediate operands'.
 *
 * If the method is abstract (from an abstract class, or from an interface)
 * the result is undefined.
 */
static int get_method_length(jclass clazz, jmethodID method)
{
	jvmdiError err;
	jlocation start;
	jlocation end;

	err = jvmdi->GetMethodLocation(clazz, method, &start, &end);

	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
			report_jvmdi_error("Error getting method location information",
			                   err);
		}
		return -1;
	}

	if (start == -1 || end == -1) {
		return -1;
	}

	return end - start + 1;
}


/*
 * Returns 0 on success
 */
static int coverage_put(JNIEnv *env, jclass clazz, jmethodID method,
                         coverage *cov)
{
	int result;
	char *key = make_method_key(env, clazz, method, ":coverage");

	if (!key) return 1;

	result = hash_put(hash, key, coverage_sizeof(cov), (void *)cov);

	free(key);

	return result;
}


void create_initial_coverage_data(JNIEnv *env, jclass clazz, jmethodID method)
{
	coverage *cov;

	int bitset_size = get_method_length(clazz, method);
	cov = coverage_allocate(bitset_size);
	if (!cov) return;

	coverage_put(env, clazz, method, cov);

	coverage_free(cov);
}


/*
 * Returns the coverage bitmap for the given method, or null if there is a
 * problem, or the method isn't being profiled.
 * 
 * The returned structure should not be freed.
 */
static coverage *coverage_get(JNIEnv *env, jclass clazz, jmethodID method)
{
	coverage *result;
	int size;
	char *key;

	key = make_method_key(env, clazz, method, ":coverage");

	if (!key) {
		return NULL;
	}

	if (hash_get(hash, key, &size, (void **)&result)) {
		result = NULL;
	}

	free(key);

	return result;
}



/*
jint last_clazz = NULL;
jmethodID last_method = NULL;
*/

void update_coverage(JNIEnv *env, JVMDI_single_step_event_data *event)
{
	jclass clazz = event->clazz;
	jmethodID method = event->method;
/*	jvmdiError err;
	jint clazz_code;

	err = jvmdi->GetObjectHashCode(clazz, &clazz_code);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("GetObjectHashCode failed", err);
	}

	if (last_clazz==clazz_code && last_method==method) {
		EPRINT("HIT");
	} else {
		EPRINT("miss");
	}

	last_clazz = clazz_code;
	last_method = method;
*/	
	coverage *cov = coverage_get(env, clazz, method);

	if (cov) {
		coverage_mark(cov, event->location);
		coverage_put(env, clazz, method, cov);
	}
}
