/*
 * cvprof.c
 * $Revision: 1.9 $
 *
 * Copyright (c) 2004 David Holroyd
 * See the file COPYING for license terms
 *
 * http://www.badgers-in-foil.co.uk/projects/cvprof/
 */

#include <stdlib.h>
#include <string.h>

/*
 * jni_md.h assumes visual studio '__int64' will be available under windows.
 * typedef it to gcc equivalent, 'long long'
 */
#ifndef __int64
#  ifdef __GNUC__
     typedef long long __int64;
#  endif
#endif

#define NO_JVMDI_MACROS
#include <jvmdi.h>
#include <jni.h>

#include "hash.h"
#include "hash_utils.h"
#include "exclusions.h"
#include "options.h"
#include "err.h"
#include "coverage.h"

#define PROFILER_UNINITIALISED 0
#define PROFILER_INITIALISED 1
#define PROFILER_SHUTDOWN -1

/* for SetEventNotificationMode's 'thread' argument */
#define ALL_THREADS NULL

#define DEFAULT_DB "coverage"

#define ACC_ABSTRACT 0x0400

typedef struct {
	int line_count;
	JVMDI_line_number_entry lines[0];
} lineinfo;


typedef struct {
	int bytecode_count;
	jbyte bytecodes[0];
} bytecodeinfo;


/* Interface to the Java debug sub-system */
JVMDI_Interface_1 *jvmdi;

/* Flag to determine initialisation has occurred */
int init = PROFILER_UNINITIALISED;

/* Used to control muti-threaded access to profile data */
JVMDI_RawMonitor profile_lock;


exclusion_list *exclude_packages = NULL;

/* options -- constructed from command line args */
options opts;

/* Handle on the profile data itself */
hash_table hash;


/*
 * TODO
 *  - compile this with gprof instrumentation
 *  - probably cache the hash datums, and only write if thisMethod != lastMethod
 *  - look at alternatives to dbm (mmap?)
 *  - record bytecode data so that we can tell which positions in a coverage
 *    bitmap do not actualy mark the position of a bytecode
 */

static inline void profile_lock_enter()
{
	jvmdiError err;
	err = jvmdi->RawMonitorEnter(profile_lock);
	if (err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't aquire monitor", err);
		exit(-2);
	}
}


static inline void profile_lock_exit()
{
	jvmdiError err;
	err = jvmdi->RawMonitorExit(profile_lock);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't release monitor", err);
	}
}


/*
 * Returns 0 on success
 */
static int setup_package_exclusion_list()
{
	if (opts.exclusions_file == NULL) {
		exclusion_list_add(&exclude_packages, strdup("Lsun/"));
		exclusion_list_add(&exclude_packages, strdup("Ljava/"));
		exclusion_list_add(&exclude_packages, strdup("Ljavax/"));
		exclusion_list_add(&exclude_packages,strdup("Lcom/sun/"));
		return 0;
	} else {
		FILE *in = fopen(opts.exclusions_file, "r");
		if (!in) {
			EPRINT("Can't open exclusions file \"%s\"",
			       opts.exclusions_file);
			return 1;
		}
		exclude_packages = exclusion_list_read(in);
		fclose(in);

		return 0;
	}
}

static inline int is_array_signature(const char *sig)
{
	return sig[0] == '[';
}

static int should_profile_class(jclass clazz)
{
	jvmdiError err;
	jboolean is_interface;
	char *clazz_sig;
	int result = 1;

	err = jvmdi->IsInterface(clazz, &is_interface);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't determine if class is an interface", err);
		return 0;
	}

	if (is_interface) {
		return 0;
	}

	err = jvmdi->GetClassSignature(clazz, &clazz_sig);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't get class details", err);
		return 0;
	}

	if (is_array_signature(clazz_sig)
	    || exclusion_list_matches(exclude_packages, clazz_sig)) {
		result = 0;
	}
	jvmdi->Deallocate((jbyte *)clazz_sig);

	return result;
}


static int should_profile_method(jclass clazz, jmethodID method)
{
	jvmdiError err;
	jboolean method_type;
	jint modifiers;

	err = jvmdi->GetMethodModifiers(clazz, method, &modifiers);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("GetMethodModifiers failed", err);
		return 0;
	}
	/* TODO: do we need to check for 'native' too? */
	if (modifiers & ACC_ABSTRACT) {
		return 0;
	}

	err = jvmdi->IsMethodSynthetic(clazz, method, &method_type);
	if(err != JVMDI_ERROR_NONE) {
		/*
		 * JDK1.2 doesn't allow us to test for synthetic methods.
		 * Don't return false, or nothing can be profiled.
		 */
		if (err != JVMDI_ERROR_NOT_IMPLEMENTED) {
			report_jvmdi_error("IsMethodSynthetic failed", err);
			return 0;
	  	}
	} else if (method_type) {
		return 0;
	}

	err = jvmdi->IsMethodNative(clazz, method, &method_type);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("IsMethodNative failed", err);
		return 0;
	}
	if (method_type) return 0;

	return 1;
}


static int should_profile_thread(JNIEnv *env, jthread thread)
{
	/*
	 * TODO: test for threads (probably by name) that we don't need to
	 *       enable single-stepping/profiling for, e.g. "Signal Dispatcher",
	 *       "process reaper", "SIGINT handler"
	 */
	return 1;
}


static void record_source_file_name(JNIEnv *env, jclass clazz,
                                    const char *source_file)
{
	char *key = make_class_key(env, clazz, ":source");

	if (key) {
		hash_put(hash, key, strlen(source_file), source_file);
		free(key);
	}
}


static void create_line_number_data(JNIEnv *env, jclass clazz, jmethodID method)
{
	jvmdiError err;
	jint line_number_count;
	JVMDI_line_number_entry *line_number_table;

	int table_size;
	lineinfo *table_copy;
	char *key;

	err = jvmdi->GetLineNumberTable(clazz, method,
	                                &line_number_count,
	                                &line_number_table);
	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
			report_jvmdi_error("Error retriving line information",
			                   err);
		}
		/* else, Maybe native method or no debug. Ignore */
		return;
	}

	key = make_method_key(env, clazz, method, ":lines");
	if (key) {
		table_size = sizeof(lineinfo) + sizeof(JVMDI_line_number_entry)
		            * line_number_count;

		table_copy = (lineinfo *)malloc(table_size);
		if (table_copy) {
			table_copy->line_count = line_number_count;
			memcpy(table_copy->lines,
			       line_number_table,
			   sizeof(JVMDI_line_number_entry) * line_number_count);

			if (hash_put(hash, key, table_size, table_copy)){
				EPRINT("Writing line number data to db failed");
			}
			free(table_copy);
		}
		free(key);
	}

	jvmdi->Deallocate((jbyte *)line_number_table);
}


static void create_bytecode_data(JNIEnv *env, jclass clazz,
                                         jmethodID method)
{
	jvmdiError err;
	jint bytecode_count;
	jbyte *bytecodes;
	char *key;

	err = jvmdi->GetBytecodes(clazz, method,
	                                 &bytecode_count, &bytecodes);
	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_NOT_IMPLEMENTED) {
			report_jvmdi_error("Couldn't get bytecodes", err);
		}
		return;
	}

	key = make_method_key(env, clazz, method, ":bytecode");
	if (key) {
		int bytecode_size = sizeof(bytecodeinfo)
		                  + sizeof(jbyte)*bytecode_count;
		bytecodeinfo *value = (bytecodeinfo *)malloc(bytecode_size);
		value->bytecode_count = bytecode_count;
		memcpy(value->bytecodes, bytecodes,
		       sizeof(jbyte) * bytecode_count);

		if (hash_put(hash, key, bytecode_size, value)) {
			EPRINT("Writing line number data to db failed");
		}
	}

	jvmdi->Deallocate(bytecodes);
}


static void create_method_data(JNIEnv *env, jclass clazz, jmethodID method)
{
	create_line_number_data(env, clazz, method);

	create_bytecode_data(env, clazz, method);

	create_initial_coverage_data(env, clazz, method);
}


static void create_class_data(JNIEnv *env, jclass clazz)
{
	jvmdiError err;
	char *source_file;
	jmethodID *methods;
	jint methodCount;
	int i;

	err = jvmdi->GetSourceFileName(clazz, &source_file);
	if (err == JVMDI_ERROR_NONE) {
		record_source_file_name(env, clazz, source_file);
		jvmdi->Deallocate((jbyte *)source_file);
	} else if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
		report_jvmdi_error("GetSourceFileName failed", err);
	}

	err = jvmdi->GetClassMethods(clazz, &methodCount, &methods);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Couldn't get class methods", err);
		return;
	}

	for(i = 0 ; i<methodCount ; i++) {
		jmethodID method = methods[i];
		if (should_profile_method(clazz, method)) {
			create_method_data(env, clazz, method);
		}
	}
	jvmdi->Deallocate((jbyte *)methods);
}

/*
 * In JDK1.2, lots of class-prepare events occur before the VM-init event, so
 * we miss the chance to record those classes' data.  This method is used to
 * initialize data structures for any of those classes we're to profile.
 */
static void create_already_loaded_class_data(JNIEnv *env)
{
	jvmdiError err;
	jint class_count;
	jclass *classes;
	int i;
	
	err = jvmdi->GetLoadedClasses(&class_count, &classes);
	if (err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Couldn't get class methods", err);
		return;
	}
	
	for (i=0; i<class_count; i++) {
		jclass clazz = classes[i];
		if (should_profile_class(clazz)) {
			create_class_data(env, clazz);
		}
		(*env)->DeleteGlobalRef(env, clazz);
	}
	jvmdi->Deallocate((jbyte *)classes);
}

static inline void handle_vm_init(JNIEnv *env)
{
	if(jvmdi->CreateRawMonitor("Profile", &profile_lock)!=JVMDI_ERROR_NONE){
		EPRINT("Can't create raw monitor lock");
	}

	init = PROFILER_INITIALISED;

	profile_lock_enter();
	create_already_loaded_class_data(env);
	profile_lock_exit();
}


static inline void handle_vm_death(JNIEnv *env)
{
	/* try to make sure nothing holds the lock when we destry it */
	init = PROFILER_SHUTDOWN;
	profile_lock_enter();
	profile_lock_exit();

	jvmdi->DestroyRawMonitor(profile_lock);
	hash_close(hash);
	exclusion_list_free(exclude_packages);
	options_free(&opts);
}


static inline void handle_single_step(JNIEnv *env,
                                      JVMDI_single_step_event_data *event)
{
	profile_lock_enter();
	update_coverage(env, event);
	profile_lock_exit();
}


static inline void handle_class_prepare(JNIEnv *env,
                                        JVMDI_class_event_data *event)
{
	jclass clazz = event->clazz;
	if (should_profile_class(clazz)) {
		profile_lock_enter();
		create_class_data(env, clazz);
		profile_lock_exit();
	}
}

static inline void handle_thread_start(JNIEnv *env,
                                       JVMDI_thread_change_event_data *event)
{
	jvmdiError err;

	if (should_profile_thread(env, event->thread)) {
		err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
		                                      JVMDI_EVENT_SINGLE_STEP,
		                                      event->thread);
		if(err != JVMDI_ERROR_NONE)
			report_jvmdi_error("Can't enable single step", err);
	}
}

/*
 * If we don't disable single-step events, the Windows 1.3 JVM hangs at exit,
 * so this handler disables notification for each thread as it ends.  This
 * handle may get called after JVMDI_EVENT_VM_DEATH is recieved, so care must
 * be taken not to touch any resources (e.g. the lock) that that event handler
 * releases.
 */
static inline void handle_thread_end(JNIEnv *env,
                                       JVMDI_thread_change_event_data *event)
{
	jvmdiError err;
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_SINGLE_STEP,
	                                      event->thread);
	if(err != JVMDI_ERROR_NONE)
		report_jvmdi_error("Can't enable single step", err);
}


static inline int should_ignore_event(JVMDI_Event *event)
{
	/* Allow THREAD_END after VM_DEATH so we can clean up */
	if(init!=PROFILER_INITIALISED
	   && event->kind!=JVMDI_EVENT_VM_INIT
	   && event->kind!=JVMDI_EVENT_THREAD_END)
	{
		if(init == PROFILER_UNINITIALISED) {
			EPRINT("Unexpected event before VM init: %d",
			       (int)event->kind);
		}
		/* you can get events after VM_DEATH */
		return 1;
	}
	return 0;
}


JNIEXPORT void hook(JNIEnv *env, JVMDI_Event *event)
{
	if (should_ignore_event(event)) return;

	switch(event->kind) {
	    case JVMDI_EVENT_VM_INIT:
		handle_vm_init(env);
		break;
	    case JVMDI_EVENT_VM_DEATH:
		handle_vm_death(env);
		break;
	    case JVMDI_EVENT_SINGLE_STEP:
		handle_single_step(env, &event->u.single_step);
		break;
	    case JVMDI_EVENT_CLASS_PREPARE:
		handle_class_prepare(env, &event->u.class_event);
		break;
	    case JVMDI_EVENT_THREAD_START:
		handle_thread_start(env, &event->u.thread_change);
		break;
	    case JVMDI_EVENT_THREAD_END:
		handle_thread_end(env, &event->u.thread_change);
		break;
	    default:
		EPRINT("Unexpected event type: %d", (int)event->kind);
	}
}


static void check_jvmdi_caps(JVMDI_capabilities *caps)
{
	if(!caps->can_get_synthetic_attribute) {
		EPRINT("Can't get synthetic attribute; profile will include synthesized methods");
	}
	if(!caps->can_get_bytecodes) {
		EPRINT("Can't get bytecodes; may be a problem once reporting tools need to use them");
	}
}


static jvmdiError enable_jvmdi_events()
{
	jvmdiError err;

	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_VM_INIT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable VM init event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_VM_DEATH,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable VM death event");

	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_THREAD_START,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable thread start event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_THREAD_END,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable thread end event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_SINGLE_STEP,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		report_jvmdi_error("Can't disable single step event", err);
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_EXCEPTION_CATCH,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable exception catch event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_BREAKPOINT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable breakpoint event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FIELD_MODIFICATION,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable field modification event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FIELD_ACCESS,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable field access event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FRAME_POP,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable frame pop event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_METHOD_ENTRY,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable method entry event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_METHOD_EXIT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable method exit event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_EXCEPTION,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable exception event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_USER_DEFINED,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable user defined event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_CLASS_UNLOAD,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable class unload event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_CLASS_LOAD,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable class load event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_CLASS_PREPARE,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable class prepare event");

	return jvmdi->SetEventHook(hook);
}


JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *jvm, void *options, void *reserved)
{
	int jvm_err;
	jvmdiError err;
	JVMDI_capabilities caps;

	jvm_err = (*jvm)->GetEnv(jvm, (void **)&jvmdi, JVMDI_VERSION_1);
	if(jvm_err == JNI_EDETACHED) {
		EPRINT("VM Detached");
		return JNI_ERR;
	}
	if(jvm_err == JNI_EVERSION) {
		EPRINT("Couldn't obtain JVMDI version 1 interface");
		return JNI_ERR;
	}

	if (options_process_list(&opts, (char *)options)) {
		return JNI_EINVAL;
	}

	if (setup_package_exclusion_list()) {
		return JNI_ERR;
	}

	err = jvmdi->GetCapabilities(&caps);
	if(err != JVMDI_ERROR_NONE) {
		if(err == JVMDI_ERROR_ACCESS_DENIED) {
			EPRINT("Profile can't be started since debugging is disabled. Specify -Xdebug on JVM command line (SUN JVM)");
			return JNI_ERR;
		}
		report_jvmdi_error("Can't get capabilities", err);
	}
	check_jvmdi_caps(&caps);

	if (enable_jvmdi_events() != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't set event hook", err);
		return JNI_ERR;
	}

	hash = hash_open(DEFAULT_DB);
	if (!hash) {
		return JNI_ERR;
	}

	return JNI_OK;
}
